/* automatically generated from ./terminal-window-ui.xml */
#ifdef __SUNPRO_C
#pragma align 4 (terminal_window_ui)
#endif
#ifdef __GNUC__
static const char terminal_window_ui[] __attribute__ ((__aligned__ (4))) =
#else
static const char terminal_window_ui[] =
#endif
{
  "<ui><menubar name=\"main-menu\"><menu action=\"file-menu\"><menuitem ac"
  "tion=\"new-tab\"/><menuitem action=\"new-window\"/><separator/><menuite"
  "m action=\"detach-tab\"/><separator/><menuitem action=\"close-tab\"/><m"
  "enuitem action=\"close-window\"/></menu><menu action=\"edit-menu\"><men"
  "uitem action=\"copy\"/><menuitem action=\"paste\"/><menuitem action=\"p"
  "aste-selection\"/><separator/><menuitem action=\"select-all\"/><separat"
  "or/><menuitem action=\"preferences\"/></menu><menu action=\"view-menu\""
  "><menuitem action=\"show-menubar\"/><menuitem action=\"show-toolbar\"/>"
  "<menuitem action=\"show-borders\"/><menuitem action=\"fullscreen\"/></m"
  "enu><menu action=\"terminal-menu\"><menuitem action=\"set-title\"/><sep"
  "arator/><menuitem action=\"search\"/><menuitem action=\"search-next\"/>"
  "<menuitem action=\"search-prev\"/><separator/><menuitem action=\"set-en"
  "coding\"/><separator/><menuitem action=\"reset\"/><menuitem action=\"re"
  "set-and-clear\"/></menu><menu action=\"tabs-menu\"><menuitem action=\"p"
  "rev-tab\"/><menuitem action=\"next-tab\"/><separator/><menuitem action="
  "\"move-tab-left\"/><menuitem action=\"move-tab-right\"/><separator/><pl"
  "aceholder name=\"placeholder-tab-items\"/></menu><menu action=\"help-me"
  "nu\"><menuitem action=\"contents\"/><menuitem action=\"about\"/></menu>"
  "</menubar><popup name=\"popup-menu\"><menuitem action=\"new-window\"/><"
  "menuitem action=\"new-tab\"/><separator/><menuitem action=\"copy\"/><me"
  "nuitem action=\"paste\"/><separator/><menuitem action=\"show-menubar\"/"
  "><menuitem action=\"fullscreen\"/><menuitem action=\"preferences\"/><se"
  "parator/><menuitem action=\"input-methods\"/></popup><popup name=\"tab-"
  "menu\"><menuitem action=\"detach-tab\"/><menuitem action=\"set-title\"/"
  "><separator/><menuitem action=\"move-tab-left\"/><menuitem action=\"mov"
  "e-tab-right\"/><separator/><menu action=\"tabs-menu\"><placeholder name"
  "=\"placeholder-tab-items\"/></menu><separator/><menuitem action=\"close"
  "-tab\"/></popup><toolbar name=\"main-toolbar\"><toolitem action=\"new-t"
  "ab\"/><toolitem action=\"new-window\"/><separator/><toolitem action=\"c"
  "opy\"/><toolitem action=\"paste\"/><separator/><toolitem action=\"searc"
  "h\"/><separator/><toolitem action=\"fullscreen\"/><toolitem action=\"pr"
  "eferences\"/><separator/><toolitem action=\"prev-tab\"/><toolitem actio"
  "n=\"next-tab\"/><separator/><toolitem action=\"contents\"/></toolbar></"
  "ui>"
};

static const unsigned terminal_window_ui_length = 2220u;

